import functools
import signal
import sys
import requests
from requests.adapters import HTTPAdapter
from os import getenv
# noinspection PyUnresolvedReferences
from requests.packages.urllib3.util.retry import Retry
from .__logger__ import get_logger


logger = get_logger()


# pylint: disable=R0902
class EurekaService:
    """
    Class for subscribe or annotate EurekaService

    - **parameters**, **types** ::

        :property name: name of service. Can't be changed after __init__
        :type name: str
        :property host: service address (listener)
        :type env: str
        :property env: type of environment. Can't be changed after __init__
        :type uri: str
        :property uri: URL to Eureka. Can't be changed after __init__
        :type host: str
        :property port: service port (listener)
        :type port: int
        :property instance_data: read-only dictionary with instance description
        :property status: returns status of registration
        :type status: bool


    :Example:

    - register own service

        IP_OF_MY_SERVICE = '10.10.10.10'
        PORT_OF_MY_SERVICE = 9000
        EURIKA_URL = "http://eureka.domain"

        service = EurekaService(
            "MYSERVICE",
            "dev",
            EURIKA_URL,
            IP_OF_MY_SERVICE,
            PORT_OF_MY_SERVICE,
            # auto_close=True if service must be unregistred on SIGINT
        )
        service.register()
        service.setup_sigint_callback()

        ...
        # or
        service.unregister()


    - subscribe to existing service

        service = EurekaService("MY-RDBMS", 'dev', EURIKA_URL)
        service.get_connection()
        db_host, db_port = service.connection

    - get status of service

        current_instance_registred = service.status

    """

    __registered = False
    MSG_NOT_REGISTRED = "Instance doesn't registered from this instance. "

    def __init__(
            self,
            name,
            env,
            uri,
            **kwargs
    ):
        """
        :param uri: URL to Eureka
        :param name: name of current service
        :param host: current host
        :param port: current port
        :param env: environment variables for service
        :param auto_close: service must be unregistred on SIGINT
        """
        self.__uri = uri
        self.__name = name
        self.__host = kwargs.get('host', getenv('HOST', ''))
        self.__port = kwargs.get('port', getenv('PORT', ''))
        self.__env = kwargs.get('env', getenv('ENV', ''))

        retries = Retry(
            total=5,
            backoff_factor=0.1
        )

        self.__req = session = requests.session()
        session.headers.update({
            'Accept': 'application/json',
            'Content-Type': 'application/json',
        })
        session.mount('http://', HTTPAdapter(max_retries=retries))

        if kwargs.get('auto_close', True):
            self.setup_sigint_callback()

        if kwargs.get('is_service_provider', False):
            self.register()

    @property
    def uri(self):
        return self.__uri

    @property
    def env(self):
        return self.__env

    @property
    def name(self):
        return self.__name

    @property
    def host(self):
        return self.__host

    @property
    def port(self):
        return self.__port

    def register(self):
        """
        Method for register declared service
        """
        url = f'{self.uri}/eureka/apps/{self.name}'
        try:
            logger.info(f'REGISTER "{self.name}" @ "{self.uri}"')
            with self.__req.post(url, json=self.instance_data) as r:
                if r.status_code == 204:
                    logger.info(
                        f'Successfully registered "{self.name}" in EUREKA @ "{self.uri}"'
                    )
                    self.__registered = True
                else:
                    logger.error(f'"{self.name}" doesn\'t registred @ "{self.uri}"')
        except requests.ConnectionError as e:
            logger.error(
                f'"{self.name}" doesn\'t registred @ "{self.uri}". '
                f'ConnectionError: {url} \n {e}'
            )

    def heartbeat(self):
        if not self.status:
            raise LookupError(self.MSG_NOT_REGISTRED)
        instance_id = self.instance_data['instance']['instanceId']
        url = f'{self.uri}/eureka/apps/{self.name}/{instance_id}'
        try:
            with self.__req.put(url) as resp:
                if resp.ok:
                    logger.info(f'Hearbeat confirmed for "{self.name}" at "{self.uri}"')
                else:
                    logger.error(
                        f'Hearbeat not confirmed for "{self.name}" at "{self.uri}". '
                        f'Unexpected response received.'
                    )
                    raise RuntimeError
        except requests.ConnectionError:
            logger.error(
                f'Hearbeat not confirmed for "{self.name}" at "{self.uri}". '
                f'Connection error to "{url}".'
            )
            raise RuntimeError

    def _unregister(self):
        """
        "Unsafe" unregister
        """
        instance_id = self.instance_data['instance']['instanceId']
        url = f'{self.uri}/eureka/apps/{self.name}/{instance_id}'
        try:
            logger.info(f'UNREGISTER "{self.name}" @ "{url}"')
            response = requests.delete(url)
            if response.status_code == 200:
                logger.info(f'Successfully unregistered "{self.name}" @ "{self.uri}"')
                self.__registered = False
        except requests.ConnectionError:
            logger.error(f'ConnectionError: "{url}"')

    def unregister(self):
        """
        "Safe" unregister
        """
        if self.__registered:
            self._unregister()
        else:
            raise LookupError(self.MSG_NOT_REGISTRED)

    force_unregister = _unregister

    def get_existing(self):
        """
        Get config for service from eurika
        """
        url = f'{self.uri}/eureka/apps/{self.name}'

        ip, port = None, None
        with self.__req.get(url) as r:
            if r.ok:
                print(r.content)
                json_response = r.json()
                application = json_response.get('application', {})
                for instance in application.get('instance', []):
                    service_status = instance['status']
                    if service_status == 'UP':
                        ip, *_ = instance['hostName'].split(':')
                        port = instance['port']['$']
                        break
        self.__host, self.__port = ip, port

    @property
    def instance_data(self):
        return {
            "instance": {
                "instanceId": f"{self.name}.{self.env}:{self.name}:{self.host}:{self.port}",
                "hostName": f"{self.host}:{self.port}",
                "app": f"{self.name}",
                "ipAddr": f"{self.host}",
                "vipAddress": f"{self.name}",
                "secureVipAddress": f"{self.name}",
                "port": {
                    "$": f"{self.port}",
                    "@enabled": "true"
                },
                "securePort": {
                    "$": "443",
                    "@enabled": "false"
                },
                "homePageUrl": f"http://{self.host}:{self.port}/",
                "statusPageUrl": f"http://{self.host}:{self.port}/info/",
                "healthCheckUrl": f"http://{self.host}:{self.port}/health/",
                "dataCenterInfo": {
                    "@class": "com.netflix.appinfo.MyDataCenterInfo",
                    "name": "MyOwn"
                },
                "leaseInfo": {
                    "evictionDurationInSecs": "90"
                },
                "metadata": {
                    "management.port": f"{self.port}"
                }
            }
        }

    @property
    def status(self):
        """
        :return: is_registered
        """
        return self.__registered

    @property
    def connection(self):
        return self.host, self.port

    def setup_sigint_callback(self):
        """
        Callback for unregister service on SIGINT
        """

        def _sigint_handler(signum, _):
            self.unregister()
            logger.info(f"\n\n CUSTOM SIGINT HANDLER {signum}\n\n")
            sys.exit(0)

        signal.signal(signal.SIGINT, _sigint_handler)

    def __del__(self):
        self.unregister()

    def __repr__(self):
        registred = ''
        if self.status:
            registred = 'registred '

        return f'<EurekaService({registred}"{self.name}"@"{self.uri}")>'

    def __str__(self):
        return self.name


RegEurekaService = functools.partial(EurekaService, is_service_provider=True)
RegEurekaService.__doc__ = EurekaService.__doc__
